import { Server } from 'socket.io'
import { TSocket, ExtendedCharacter } from '../../utilities/Types'
import ZoneManager from '../../managers/ZoneManager'
import prisma from '../../utilities/Prisma'
import { AStar } from '../../utilities/Player/AStar'
import Rotation from '../../utilities/Player/Rotation'

const moveTokens = new Map<number, symbol>()

export default function setupCharacterMove(socket: TSocket, io: Server) {
  socket.on('character:move', async ({ position_x, position_y }: { position_x: number; position_y: number }) => {
    const { character } = socket
    if (!character) return console.error('character:move error', 'Character not found')

    moveTokens.set(character.id, Symbol('moveToken'))
    const grid = await ZoneManager.getGrid(character.zoneId)
    if (!grid?.length) return console.error('character:move error', 'Grid not found or empty')

    const start = { x: Math.floor(character.position_x), y: Math.floor(character.position_y) }
    const end = { x: Math.floor(position_x), y: Math.floor(position_y) }

    if (isObstacle(end, grid)) return socket.emit('character:moveError', 'Destination is an obstacle')

    const path = AStar.findPath(start, end, grid)
    if (!path.length) return socket.emit('character:moveError', 'No valid path found')

    character.isMoving = true
    io.in(character.zoneId.toString()).emit('character:moved', character)
    moveAlongPath(socket, io, path, grid).catch(console.error)
  })
}

async function moveAlongPath(socket: TSocket, io: Server, path: Array<{ x: number; y: number }>, grid: number[][]) {
  const { character } = socket
  if (!character) return

  const moveToken = moveTokens.get(character.id)
  const stepDuration = 250
  const updateInterval = 50

  for (let i = 0; i < path.length - 1; i++) {
    const startTime = Date.now()
    while (Date.now() - startTime < stepDuration) {
      if (moveTokens.get(character.id) !== moveToken) return

      const progress = (Date.now() - startTime) / stepDuration
      const current = interpolatePosition(path[i], path[i + 1], progress)

      if (isObstacle(current, grid)) break

      await updateCharacter(character, current, Rotation.calculate(path[i].x, path[i].y, path[i + 1].x, path[i + 1].y))
      io.in(character.zoneId.toString()).emit('character:moved', character)
      await new Promise((resolve) => setTimeout(resolve, updateInterval))
    }
  }

  if (moveTokens.get(character.id) === moveToken) {
    character.isMoving = false
    moveTokens.delete(character.id)
    io.in(character.zoneId.toString()).emit('character:moved', character)
  }
}

const interpolatePosition = (start: { x: number; y: number }, end: { x: number; y: number }, progress: number) => ({
  x: start.x + (end.x - start.x) * progress,
  y: start.y + (end.y - start.y) * progress
})

const isObstacle = ({ x, y }: { x: number; y: number }, grid: number[][]) => grid[Math.floor(y)]?.[Math.floor(x)] === 1

async function updateCharacter(character: ExtendedCharacter, { x, y }: { x: number; y: number }, rotation: number) {
  Object.assign(character, { position_x: x, position_y: y, rotation })
  ZoneManager.updateCharacterInZone(character.zoneId, character)
  await prisma.character.update({
    where: { id: character.id },
    data: { position_x: x, position_y: y, rotation }
  })
}
